from abc import ABC, abstractmethod


class BaseSerializer(ABC):
    @abstractmethod
    def dumps(self, obj) -> str:
        pass

    @abstractmethod
    def loads(self, string: str):
        pass

    def dump(self, obj, source_file):
        source_file.write(self.dumps(obj))

    def load(self, source_file):
        return self.loads(source_file.read())


import types
from types import NoneType as nonetype, \
                  ModuleType as moduletype, \
                  CodeType as codetype, \
                  FunctionType as functype, \
                  BuiltinFunctionType as bldinfunctype, \
                  CellType as celltype, \
                  MappingProxyType as mapproxytype, \
                  WrapperDescriptorType as wrapdesctype, \
                  MethodDescriptorType as metdesctype, \
                  GetSetDescriptorType as getsetdesctype

CODE_PROPS = [prop.__name__ for prop in [
        codetype.co_argcount,
        codetype.co_posonlyargcount,
        codetype.co_kwonlyargcount,
        codetype.co_nlocals,
        codetype.co_stacksize,
        codetype.co_flags,
        codetype.co_code,
        codetype.co_consts,
        codetype.co_names,
        codetype.co_varnames,
        codetype.co_filename,
        codetype.co_name,
       # codetype.co_qualname,
        codetype.co_firstlineno,
        codetype.co_lnotab,
     #   codetype.co_exceptiontable,
        codetype.co_freevars,
        codetype.co_cellvars]
    ]

UNIQUE_TYPES = [
    mapproxytype,
    wrapdesctype,
    metdesctype,
    getsetdesctype,
    bldinfunctype
]


import inspect
from Serializers import nonetype, moduletype, codetype, celltype, \
                          functype, bldinfunctype, \
                          mapproxytype, wrapdesctype, metdesctype, getsetdesctype, \
                          CODE_PROPS, UNIQUE_TYPES


class DictSerializer:
    TYPE_KW = "type"
    SOURCE_KW = "source"

    CODE_KW = "__code__"
    GLOBALS_KW = functype.__globals__.__name__
    NAME_KW = "__name__"
    DEFAULTS_KW = "__defaults__"
    CLOSURE_KW = functype.__closure__.__name__

    BASES_KW = "__bases__"
    DICT_KW = "__dict__"

    CLASS_KW = "__class__"

    OBJECT_KW = "object"

    @classmethod
    def to_dict(cls, obj, is_inner_func=False):
        if type(obj) in (int, float, bool, str, nonetype):
            return obj

        if type(obj) is list:
            return [cls.to_dict(o) for o in obj]

        if type(obj) is dict:
            # Since the key in the dictionary can be a hashable object, which will be represented as a non-hashable
            # dictionary, it is easier to represent the dictionary as a list of key-value pairs
            return {cls.TYPE_KW: dict.__name__,
                    cls.SOURCE_KW: [[cls.to_dict(item[0]), cls.to_dict(item[1])] for item in obj.items()]}

        if type(obj) in (set, frozenset, tuple, bytes, bytearray):
            return {cls.TYPE_KW: type(obj).__name__,
                    cls.SOURCE_KW: cls.to_dict([*obj])}

        if type(obj) is complex:
            return {cls.TYPE_KW: complex.__name__,
                    cls.SOURCE_KW: {complex.real.__name__: obj.real,
                                    complex.imag.__name__: obj.imag}}

        if type(obj) is moduletype:
            return {cls.TYPE_KW: moduletype.__name__,
                    cls.SOURCE_KW: obj.__name__}

        if type(obj) is codetype:
            code = {cls.TYPE_KW: codetype.__name__}
            source = {}

            for (key, value) in inspect.getmembers(obj):
                if key in CODE_PROPS:
                    source[key] = cls.to_dict(value)

            code.update({cls.SOURCE_KW: source})
            return code

        if type(obj) is celltype:
            return {cls.TYPE_KW: celltype.__name__,
                    cls.SOURCE_KW: cls.to_dict(obj.cell_contents)}

        if type(obj) in (staticmethod, classmethod):
            return {cls.TYPE_KW: type(obj).__name__,
                cls.SOURCE_KW: cls.to_dict(obj.__func__, is_inner_func)}

        if inspect.isroutine(obj):
            source = {}

            # Code
            source[cls.CODE_KW] = cls.to_dict(obj.__code__)

            # Global vars
            gvars = cls.__get_gvars(obj, is_inner_func)
            source[cls.GLOBALS_KW] = cls.to_dict(gvars)

            # Name
            source[cls.NAME_KW] = cls.to_dict(obj.__name__)

            # Defaults
            source[cls.DEFAULTS_KW] = cls.to_dict(obj.__defaults__)

            # Closure
            source[cls.CLOSURE_KW] = cls.to_dict(obj.__closure__)

            return {cls.TYPE_KW: functype.__name__,
                    cls.SOURCE_KW: source}

        elif inspect.isclass(obj):
            source = {}

            # Name
            source[cls.NAME_KW] = cls.to_dict(obj.__name__)

            # Bases
            source[cls.BASES_KW] = cls.to_dict(tuple(b for b in obj.__bases__ if b != object))

            # Dict
            source[cls.DICT_KW] = cls.__get_obj_dict(obj)

            return {cls.TYPE_KW: type.__name__,
                    cls.SOURCE_KW: source}

        else:
            source = {}

            # Class
            source[cls.CLASS_KW] = cls.to_dict(obj.__class__)

            # Dict
            source[cls.DICT_KW] = cls.__get_obj_dict(obj)

            return {cls.TYPE_KW: cls.OBJECT_KW,
                    cls.SOURCE_KW: source}

    @classmethod
    def __get_gvars(cls, func, is_inner_func):
        name = func.__name__
        gvars = {}

        for gvar_name in func.__code__.co_names:
            # Separating the variables that the function needs
            if gvar_name in func.__globals__:

                # Module
                if type(func.__globals__[gvar_name]) is moduletype:
                    gvars[gvar_name] = func.__globals__[gvar_name]

                # Class
                elif inspect.isclass(func.__globals__[gvar_name]):
                    # To prevent recursion, the class in which this method is declared is replaced with the
                    # name of the class. In the future, this name will be replaced by the class type
                    c = func.__globals__[gvar_name]
                    if is_inner_func and name in c.__dict__ and func == c.__dict__[name].__func__: #!!!!
                        gvars[gvar_name] = c.__name__
                    else:
                        gvars[gvar_name] = c

                # Recursion protection
                elif gvar_name == func.__code__.co_name:
                    gvars[gvar_name] = func.__name__

                else:
                    gvars[gvar_name] = func.__globals__[gvar_name]

        return gvars

    @classmethod
    def __get_obj_dict(cls, obj):
        dct = {item[0]: item[1] for item in obj.__dict__.items()}
        dct2 = {}

        for key, value in dct.items():
            if type(value) not in UNIQUE_TYPES:
                if inspect.isroutine(value):
                    # Recursion protection
                    dct2[cls.to_dict(key)] = cls.to_dict(value, is_inner_func=True)
                else:
                    dct2[cls.to_dict(key)] = cls.to_dict(value)

        return dct2

    @classmethod
    def from_dict(cls, obj, is_dict=False):

        if is_dict:
            return {cls.from_dict(item[0]): cls.from_dict(item[1]) for item in obj}

        if type(obj) not in (dict, list):
            return obj

        elif type(obj) is list:
            return [cls.from_dict(o) for o in obj]

        else:
            obj_type = obj[cls.TYPE_KW]
            obj_source = obj[cls.SOURCE_KW]

            if obj_type == dict.__name__:
                return cls.from_dict(obj_source, is_dict=True)

            # Key - type name, value - type itself. Calling by type name returns that type.
            # This is necessary for the same creation of simple collections.
            cols_dict = {t.__name__: t for t in [set, frozenset, tuple, bytes, bytearray]}
            if obj_type in cols_dict:
                return cols_dict[obj_type](cls.from_dict(obj_source))

            if obj_type == complex.__name__:
                return obj_source[complex.real.__name__] + \
                    obj_source[complex.imag.__name__] * 1j

            if obj_type == moduletype.__name__:
                return __import__(obj_source)

            if obj_type == codetype.__name__:
                return codetype(*[cls.from_dict(obj_source[prop]) for prop in CODE_PROPS])

            if obj_type == celltype.__name__:
                return celltype(cls.from_dict(obj_source))

            if obj_type == staticmethod.__name__:
                return staticmethod(cls.from_dict(obj_source))

            if obj_type == classmethod.__name__:
                return classmethod(cls.from_dict(obj_source))

            if obj_type == functype.__name__:
                code = cls.from_dict(obj_source[cls.CODE_KW])
                gvars = cls.from_dict(obj_source[cls.GLOBALS_KW])
                name = cls.from_dict(obj_source[cls.NAME_KW])
                defaults = cls.from_dict(obj_source[cls.DEFAULTS_KW])
                closure = cls.from_dict(obj_source[cls.CLOSURE_KW])

                # If there are suitable global variables, they are replaced.
                for key in gvars:
                    if key in code.co_name and key in globals():
                        gvars[key] = globals()[key]

                func = functype(code, gvars, name, defaults, closure)

                # Restoring recursion
                if func.__name__ in gvars:
                    func.__globals__.update({func.__name__: func})

                return func

            if obj_type == type.__name__:
                name = cls.from_dict(obj_source[cls.NAME_KW])
                bases = cls.from_dict(obj_source[cls.BASES_KW])
                dct = obj_source[cls.DICT_KW]
                dct = {cls.from_dict(item[0]): cls.from_dict(item[1]) for item in dct.items()}

                cl = type(name, bases, dct)

                # Restore a reference to the current class in the nested method __globals__
                for attr in cl.__dict__.values():
                    if inspect.isroutine(attr):
                        if type(attr) in (staticmethod, classmethod):
                            fglobs = attr.__func__.__globals__
                        else:
                            fglobs = attr.__globals__

                        for gv in fglobs.keys():
                            if gv == cl.__name__:
                                fglobs[gv] = cl

                return cl

            else:
                clas = cls.from_dict(obj_source[cls.CLASS_KW])
                dct = obj_source[cls.DICT_KW]
                dct = {cls.from_dict(item[0]): cls.from_dict(item[1]) for item in dct.items()}

                o = object.__new__(clas)
                o.__dict__ = dct

                return o








from Serializers.constants import \
                           nonetype, moduletype, codetype, celltype, \
                           functype, bldinfunctype, \
                           mapproxytype, wrapdesctype, metdesctype, getsetdesctype, \
                           CODE_PROPS, UNIQUE_TYPES

from Serializers.base_serializer import BaseSerializer
from Serializers.dict_serializer import DictSerializer
from Serializers.json_serializer import JsonSerializer
from Serializers.xml_serializer import XmlSerializer
from Serializers.serializers_factory import SerializersFactory, SerializerType



import re

import regex
from Serializers import BaseSerializer
from Serializers import DictSerializer

from Serializers import nonetype


class XmlSerializer(BaseSerializer):
    NONE_LITERAL = "null"

    KEY_GROUP_NAME = "key"
    VALUE_GROUP_NAME = "value"

    XML_SCHEME_SOURCE = "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " + \
                        "xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\""

    XML_SCHEME_PATTERN = "xmlns:xsi=\"http://www\.w3\.org/2001/XMLSchema-instance\" " + \
                         "xmlns:xsd=\"http://www\.w3\.org/2001/XMLSchema\""

    ELEMENTARY_NAMES_PATTERN = "int|float|bool|str|NoneType|list|dict"

    XML_ELEMENT_PATTERN = fr"(\<(?P<{KEY_GROUP_NAME}>{ELEMENTARY_NAMES_PATTERN})\>" + \
                          fr"(?P<{VALUE_GROUP_NAME}>([^<>]*)|(?R)+)\</(?:{ELEMENTARY_NAMES_PATTERN})\>)"

    FIRST_XML_ELEMENT_PATTERN = fr"(\<(?P<{KEY_GROUP_NAME}>{ELEMENTARY_NAMES_PATTERN})\s*({XML_SCHEME_PATTERN})?\>" + \
                                fr"(?P<{VALUE_GROUP_NAME}>([^<>]*)|(?R)+)\</(?:{ELEMENTARY_NAMES_PATTERN})\>)"

    def dumps(self, obj) -> str:
        obj = DictSerializer.to_dict(obj)
        return self.__dumps_from_dict(obj, is_first=True)

    def __dumps_from_dict(self, obj, is_first=False) -> str:
        if type(obj) in (int, float, bool, nonetype):
            return self.__create_xml_element(type(obj).__name__, str(obj), is_first)

        if type(obj) is str:
            data = self.__mask_symbols(obj)
            return self.__create_xml_element(str.__name__, data, is_first)

        if type(obj) is list:
            data = ''.join([self.__dumps_from_dict(o) for o in obj])
            return self.__create_xml_element(list.__name__, data, is_first)

        if type(obj) is dict:
            data = ''.join(
                [f"{self.__dumps_from_dict(item[0])}{self.__dumps_from_dict(item[1])}" for item in obj.items()])
            return self.__create_xml_element(dict.__name__, data, is_first)

        else:
            raise ValueError

    def loads(self, string: str):
        obj = self.__loads_to_dict(string, is_first=True)
        return DictSerializer.from_dict(obj)

    def __loads_to_dict(self, string: str, is_first=False):
        string = string.strip()
        xml_element_pattern = self.FIRST_XML_ELEMENT_PATTERN if is_first else self.XML_ELEMENT_PATTERN

        match = regex.fullmatch(xml_element_pattern, string)

        if not match:
            raise ValueError

        key = match.group(self.KEY_GROUP_NAME)
        value = match.group(self.VALUE_GROUP_NAME)

        if key == int.__name__:
            return int(value)

        if key == float.__name__:
            return float(value)

        if key == bool.__name__:
            return value == str(True)

        if key == str.__name__:
            return self.__unmask_symbols(value)

        if key == nonetype.__name__:
            return None

        if key == list.__name__:
            matches = regex.findall(self.XML_ELEMENT_PATTERN, value)
            return [self.__loads_to_dict(match[0]) for match in matches]

        if key == dict.__name__:
            matches = regex.findall(self.XML_ELEMENT_PATTERN, value)
            return {self.__loads_to_dict(matches[i][0]):
                        self.__loads_to_dict(matches[i + 1][0]) for i in range(0, len(matches), 2)}
        else:
            raise ValueError

    def __create_xml_element(self, name: str, data: str, is_first=False):
        if is_first:
            return f"<{name} {self.XML_SCHEME_SOURCE}>{data}</{name}>"
        else:
            return f"<{name}>{data}</{name}>"

    @staticmethod
    def __mask_symbols(string: str) -> str:
        return string.replace('&', "&amp;").replace('<', "&lt;").replace('>', "&gt;"). \
                      replace('"', "&quot;").replace("'", "&apos;")

    @staticmethod
    def __unmask_symbols(string: str) -> str:
        return string.replace("&amp;", '&').replace("&lt;", '<').replace("&gt;", '>'). \
                      replace("&quot;", '"').replace("&apos;", "'")


import re
import regex
from Serializers import BaseSerializer
from Serializers import DictSerializer


from Serializers import nonetype


class JsonSerializer(BaseSerializer):
    INF_LITERAL = str(1E1000)
    NAN_LITERAL = str(1E1000 / 1E1000)

    TRUE_LITERAL = "true"
    FALSE_LITERAL = "false"

    NULL_LITERAL = "null"

    INT_PATTERN = fr"[+-]?\d+"
    FLOAT_PATTERN = fr"(?:[+-]?\d+(?:\.\d+)?(?:e[+-]?\d+)?|[+-]?{INF_LITERAL}\b|{NAN_LITERAL}\b)"
    BOOL_PATTERN = fr"({TRUE_LITERAL}|{FALSE_LITERAL})\b"
    STRING_PATTERN = fr"\"(?:(?:\\\")|[^\"])*\""
    NULL_PATTERN = fr"\b{NULL_LITERAL}\b"

    ELEMENTARY_TYPES_PATTERN = fr"{FLOAT_PATTERN}|{INT_PATTERN}|{BOOL_PATTERN}|{STRING_PATTERN}|{NULL_PATTERN}"

    # This regex use recursive statements to be able to capture nested lists and objects.
    ARRAY_PATTERN = r"\[(?R)?(?:,(?R))*\]"
    OBJECT_PATTERN = r"\{(?:(?R):(?R))?(?:,(?R):(?R))*\}"

    VALUE_PATTERN = fr"\s*({ELEMENTARY_TYPES_PATTERN}|" + \
                    fr"{ARRAY_PATTERN}|{OBJECT_PATTERN})\s*"

    def dumps(self, obj) -> str:
        obj = DictSerializer.to_dict(obj)
        return self.__dumps_from_dict(obj)

    def __dumps_from_dict(self, obj) -> str:
        if type(obj) in (int, float):
            return str(obj)

        if type(obj) is bool:
            return self.TRUE_LITERAL if obj else self.FALSE_LITERAL

        if type(obj) is str:
            return '"' + self.__mask_quotes(obj) + '"'

        if type(obj) is nonetype:
            return self.NULL_LITERAL

        if type(obj) is list:
            return '[' + ", ".join([self.__dumps_from_dict(item) for item in obj]) + ']'

        if type(obj) is dict:
            return '{' + ", ".join([f"{self.__dumps_from_dict(item[0])}: "
                                    f"{self.__dumps_from_dict(item[1])}" for item in obj.items()]) + '}'
        else:
            raise ValueError

    def loads(self, string: str):
        obj = self.__loads_to_dict(string)
        return DictSerializer.from_dict(obj)

    def __loads_to_dict(self, string: str):
        string = string.strip()

        # Int
        match = re.fullmatch(self.INT_PATTERN, string)
        if match:
            return int(match.group(0))

        # Float
        match = re.fullmatch(self.FLOAT_PATTERN, string)
        if match:
            return float(match.group(0))

        # Bool
        match = re.fullmatch(self.BOOL_PATTERN, string)
        if match:
            return match.group(0) == self.TRUE_LITERAL

        # Str
        match = re.fullmatch(self.STRING_PATTERN, string)
        if match:
            ans = match.group(0)
            ans = self.__unmask_quotes(ans)
            return ans[1:-1]

        # None
        match = re.fullmatch(self.NULL_PATTERN, string)
        if match:
            return None

        # List
        if string[0] == '[' and string[-1] == ']':
            string = string[1:-1]
            matches = regex.findall(self.VALUE_PATTERN, string)
            return [self.__loads_to_dict(match[0]) for match in matches]

        # Dict
        if string[0] == '{' and string[-1] == '}':
            string = string[1:-1]
            matches = regex.findall(self.VALUE_PATTERN, string)

            # Variable matches will store key-value pairs in one row. Elements with
            # even indexes are keys, those with odd indexes are values.
            return {self.__loads_to_dict(matches[i][0]):
                    self.__loads_to_dict(matches[i + 1][0]) for i in range(0, len(matches), 2)}

        else:
            raise ValueError

    @staticmethod
    def __mask_quotes(string: str) -> str:
        return string.replace('\\', "\\\\").replace('"', r"\"").replace("'", r"\'")

    @staticmethod
    def __unmask_quotes(string: str) -> str:
        return string.replace('\\\\', "\\").replace(r"\"", '"').replace(r"\'", "'")




from setuptools import setup

setup(
    name="A1nzz_serializer",
    version="1.0",
    description="module for python (de)serialization in Json and Xml",
    url="https://github.com/A1nzz/Python-Labs/tree/Lab3",
    author="Mamchenko Kirill",
    author_email="kmamcenko@gmail.com",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent"
    ],
    packages=["Serializers"],
    include_package_data=True,
    install_requires=["regex"]
)



from enum import Enum

from Serializers import XmlSerializer
from Serializers import JsonSerializer


class SerializerType(Enum):
    JSON = "json"
    XML = "xml"


class SerializersFactory:
    @staticmethod
    def create_serializer(st: SerializerType):

        if st == SerializerType.JSON:
            return JsonSerializer()

        elif st == SerializerType.XML:
            return XmlSerializer()

        else:
            raise Exception("Unknown type of serialization")
